/*
 * Copyright (C) 2011 Marvell Semiconductor, Inc
 *
 * Licensed under the MIT License
 * Alternatively, you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * See COPYING-GPL.txt for the GPL license text.
 *
 * MIT License:
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 */


#include <assert.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <ctype.h>
#include <errno.h>


#include "loki.h"
#include "fw_header.h"
#include "wamlib.h"
#include "package.h"
#include "wamlib_priv.h"


static int wc_info(int argc, char **argv);
static int wc_events(int argc, char **argv);
static int wc_ping(int argc, char **argv);
static int wc_reset(int argc, char **argv);
static int wc_hardreset(int argc, char **argv);
static int wc_ssd_format(int argc, char **argv);
static int wc_acquit(int argc, char **argv);
static int wc_list_errs(int argc, char **argv);
static int wc_ssd_update_raw(int argc, char **argv);
static int wc_update_eeprom_raw(int argc, char **argv);
static int wc_read_eeprom(int argc, char **argv);
static int wc_regdump(int argc, char **argv);
static int wc_scap_calib(int argc, char **argv);
static int wc_reg_rw(int argc, char **argv);
static int wc_nvsram_dump(int argc, char **argv);

static int wc_update(int argc, char **argv);
static int wc_board_id(int argc, char **argv);
static int wc_print_hidden(int argc, char **argv);
static int wc_print_vpd(int argc, char **argv);
static int wc_convert_vpd(int argc, char **argv);
static int wc_prog_vpd(int argc, char **argv);


#define WAM_HIDDEN_HELP_STR		"-H"
#define WAM_UPDATE_MAX                  (5)

/** \wc_update_fn
 */
typedef int
(*wc_update_fn)(int fd, char *filename, int force);


/** \typedef wc_update_cmd_t
 *
 */
typedef struct wc_update_cmd wc_update_cmd_t;
struct wc_update_cmd {
        char            *u_long_cmd;	/**< long command               */
        char            *u_short_cmd;	/**< short command              */
        wc_update_fn   u_update_fn;	/**< update function            */
};


typedef struct {
        char    *long_cmd;
        char    *short_cmd;
        char    *desc;
        char    *usage;
        int     (*do_cmd)(int, char **);
        u32_t   min_args;
        u32_t   max_args;
	int	cmd_hidden;
} cmd_option_t;


const char *event_str[] = {
	"",
	"DRAM ECC Error",
	"FW Crashed",
	"SSD Not Detected",
	"SSD Not Responding",
	"SSD IO Fails",
	"SSD End of Life Reached",
	"Supercap Not Connected",
	"Supercap Voltage Low",
	"Supercap Degraded",
	"Memory Test Failed",
	"Data Backup Failed",
	"Data Restore Failed",
	"Data Loss Detected",

	"", "", "", "",
	"", "", "", "",
	"", "", "", "",
	"", "", "", "",
	"", "",

	"Device Write Ready",
	"Backup/Restore Error was cleared",
	"SSD FW updated",
	"SSD Formatted",

	"Power Loss Detected",
	"Power On Detected",
	"PCI-E Reset",
	"User/FW triggered CPU Reset",

	"Data Backup Started",
	"Data Backup Completed",
	"Data Restore Started",
	"Data Restore Completed",

	"User enabled forced ECC option",
	"User disabled forced ECC option",
	"FW Bootup",
	"Supercap Voltage",

	"Host DMA Initialized",
	"Memory Test Started",
	"Memory Test Completed",
	"Error Injected",
	"Supercap Health",
	"Supercap Health Check Started",
	"VPD Not Programmed",

	"Unknown Event"
};

char * scap_status[] = {
	"Not connected",
	"Charging",
	"Charging complete",
	"Discharging",
	"Degraded",
};

char * ssd_status[] = {
	"Not connected",
	"Erase in progress",
	"Connected",
	"Not responding",
	"IO Failing",
	"Erase pending",
	"End Of Life Reached",
};

char * memtest_status[] = {
	"Skipped",
	"Passed",
	"Failed"
};

const char * error_type[] = {
	"",
	"DMA Error",
	"SSD Write Error",
	"SSD Read Error",
};

static cmd_option_t g_cmds[] = {
	{
		.long_cmd	= "info",
		.short_cmd	= "I",
		.desc		= "Show device information",
		.usage		= "",
		.do_cmd		= wc_info,
		.min_args	= 0,
		.max_args	= 1,
	},
	{
		.long_cmd	= "events",
		.short_cmd	= "E",
		.desc		= "Show device event log. Option "
					"'clear' will show and clear the log",
		.usage		= "[clear]",
		.do_cmd		= wc_events,
		.min_args	= 0,
		.max_args	= 1,
	},
	{
		.long_cmd	= "update",
		.short_cmd	= "U",
		.desc		= "Update device firmware package",
		.usage		= "package|pkg=<package file>",
		.do_cmd		= wc_update,
		.min_args	= 1,
		.max_args	= 2,
	},
	{
		.long_cmd	= "ping",
		.short_cmd	= "p",
		.desc		= "Ping device for responsiveness",
		.usage		= "",
		.do_cmd		= wc_ping,
		.min_args	= 0,
		.max_args	= 0,
	},
	{
		.long_cmd	= "reset",
		.short_cmd	= "R",
		.desc		= "Reset device firmware",
		.usage		= "",
		.do_cmd		= wc_reset,
		.min_args	= 0,
		.max_args	= 0,
	},
	{
		.long_cmd	= "hardreset",
		.short_cmd	= "HR",
		.desc		= "Hard reset",
		.usage		= "",
		.do_cmd		= wc_hardreset,
		.min_args	= 0,
		.max_args	= 0,
	},
	{
		.long_cmd	= "list_errors",
		.short_cmd	= "L",
		.desc		= "List error blocks log "
					"(errors that occured during backup/"
					"restore)",
		.usage		= "",
		.do_cmd		= wc_list_errs,
		.min_args	= 0,
		.max_args	= 0,
	},
	{
		.long_cmd	= "clear_error",
		.short_cmd	= "C",
		.desc		= "Clear device error log (latched error "
					"needs to be cleared by user)",
		.usage		= "",
		.do_cmd		= wc_acquit,
		.min_args	= 0,
		.max_args	= 0,
	},
	{
		.long_cmd	= "update_ssd",
		.short_cmd	= "us",
		.desc		= "Update ssd firmware",
		.usage		= "<ssd firmware file>",
		.do_cmd		= wc_ssd_update_raw,
		.min_args	= 1,
		.max_args	= 1,
		.cmd_hidden	= 1,
	},
	{
		.long_cmd	= "format_ssd",
		.short_cmd	= "fssd",
		.desc		= "Format ssd attached to the device",
		.usage		= "",
		.do_cmd		= wc_ssd_format,
		.min_args	= 0,
		.max_args	= 0,
	},
	{
		.long_cmd	= "scap_calib",
		.short_cmd	= "calib",
		.desc		= "Start supercap health measurement. With the "
					"optional argument,\n\tset the "
					"automatic health check schedule in days",
		.usage		= "[value 1 to 30]",
		.do_cmd		= wc_scap_calib,
		.min_args	= 0,
		.max_args	= 1,
	},
	{
		.long_cmd	= "update_eeprom",
		.short_cmd	= "ue",
		.desc		= "Update the device eeprom",
		.usage		= "<eeprom binary file>",
		.do_cmd		= wc_update_eeprom_raw,
		.min_args	= 1,
		.max_args	= 1,
		.cmd_hidden	= 1,
	},
	{
		.long_cmd	= "read_eeprom",
		.short_cmd	= "re",
		.desc		= "Read the device eeprom and dump to console "
					"(in binary format)",
		.usage		= " > filename",
		.do_cmd		= wc_read_eeprom,
		.min_args	= 0,
		.max_args	= 0,
		.cmd_hidden	= 1,
	},
	{
		.long_cmd	= "regdump",
		.short_cmd	= "rdmp",
		.desc		= "Print device registers",
		.usage		= "",
		.do_cmd		= wc_regdump,
		.min_args	= 0,
		.max_args	= 0,
		.cmd_hidden	= 1,
	},
	{
		.long_cmd	= "reg_rw",
		.short_cmd	= "i",
		.desc		= "Read/Write register",
		.usage		= "<reg address> [value to write in hex]",
		.do_cmd		= wc_reg_rw,
		.min_args	= 1,
		.max_args	= 2,
		.cmd_hidden	= 1,
	},
	{
		.long_cmd	= "boardid",
		.short_cmd	= "bid",
		.desc		= "Identify the board ID/revision",
		.usage		= "",
		.do_cmd		= wc_board_id,
		.min_args	= 0,
		.max_args	= 0,
		.cmd_hidden	= 0,
	},
	{
		.long_cmd	= "hidden",
		.short_cmd	= "H",
		.desc		= "print all/hidden commands",
		.usage		= "",
		.do_cmd		= wc_print_hidden,
		.min_args	= 0,
		.max_args	= 0,
		.cmd_hidden	= 1,
	},
	{
		.long_cmd	= "vpd",
		.short_cmd	= "vpd",
		.desc		= "Print Vital Product Data(VPD) info",
		.usage		= "",
		.do_cmd		= wc_print_vpd,
		.min_args	= 0,
		.max_args	= 1,
	},
	{
		.long_cmd	= "convert_vpd",
		.short_cmd	= "cvpd",
		.desc		= "Convert the vpd to V2",
		.usage		= "",
		.do_cmd		= wc_convert_vpd,
		.min_args	= 0,
		.max_args	= 0,
		.cmd_hidden	= 1,
	},
	{
		.long_cmd	= "prog_vpd",
		.short_cmd	= "pvpd",
		.desc		= "Program wam vpd",
		.usage		= "\n\t\tBOARDV=<major.minor>\n"
				  "\t\tBN=<8 bytes batch number>\n"
				  "\t\tSN=<serial number>\n"
				  "\t\tWWN=<wwn> <part number>\n"
				  "\t\tPN=<part number>\n",
		.do_cmd		= wc_prog_vpd,
		.min_args	= 5,
		.max_args	= 5,
		.cmd_hidden	= 1,
	},
	{
		.long_cmd	= "nvsdump",
		.short_cmd	= "nvs",
		.desc		= "Dump internal nvram to stdout",
		.usage		= "> nvsfile.bin",
		.do_cmd		= wc_nvsram_dump,
		.min_args	= 0,
		.max_args	= 0,
		.cmd_hidden	= 1,
	},
};

static int g_nr_cmds = sizeof (g_cmds) / sizeof (cmd_option_t);
static int g_devfd;
static char *g_cmd_string;

static void
print_cmd_usage(u32_t cmd_idx, int print_hidden)
{
	cmd_option_t	*cmd = &g_cmds[cmd_idx];

	if (cmd->cmd_hidden && !print_hidden) {
		return;
	}

	printf("\n\t%s|%s %s\n\t- %s\n", cmd->long_cmd, cmd->short_cmd,
			cmd->usage, cmd->desc);
}


static void
print_usage(int print_hidden)
{
	u32_t i;

	printf("Usage: %s <device file> <cmd> <arguments>\n", g_cmd_string);
	printf("Commands:");

	for (i = 0; i < g_nr_cmds; i++)
		print_cmd_usage(i, print_hidden);
}

static int
wc_print_hidden(int argc, char **argv)
{
	print_usage(1);
	return (0);
}

static int
check_cmd(u32_t idx, char *cmd, int argc)
{
	if (strncmp(cmd, g_cmds[idx].long_cmd, 64) &&
			strncmp(cmd, g_cmds[idx].short_cmd, 64)) {
		return (-1);
	}

	if (argc - 1 < g_cmds[idx].min_args) {
		fprintf(stderr, "Too few arguments for %s\n", cmd);
		print_cmd_usage(idx, 1);
		exit(1);
	}

	if (argc - 1 > g_cmds[idx].max_args) {
		fprintf(stderr, "Too many arguments for %s\n", cmd);
		print_cmd_usage(idx, 1);
		exit(1);
	}

	return (0);
}

static int
open_dev_file(char *devfile, int *devfd)
{
	int ret;
	struct stat devstat;

	memset(&devstat, 0, sizeof(struct stat));
	ret = stat(devfile, &devstat);
	if (ret != 0) {
		fprintf(stderr, "Cannot stat file %s\n", devfile);
		perror("stat");
		return ret;
	}

	if (!S_ISCHR(devstat.st_mode)) {
		fprintf(stderr, "File %s is not a character device\n",
				devfile);
		return -1;
	}

	*devfd = open(devfile, O_RDWR);
	if (*devfd < 0) {
		fprintf(stderr, "Failed to open device: %s\n", devfile);
		perror("open");
		return -1;
	}

	return 0;
}

int
main(int argc, char **argv)
{
	u32_t	i;
	int	ret = 0;

	g_cmd_string = argv[0];

	if (argc < 3) {
		print_usage(0);
		return -1;
	}

	if (open_dev_file(argv[1], &g_devfd) != 0)
		return -1;

	argc -= 2;
	argv += 2;

	for (i = 0; i < g_nr_cmds; i++) {
		if (check_cmd(i, argv[0], argc) == 0) {
			argc--;
			argv++;
			ret = g_cmds[i].do_cmd(argc, argv);
			break;
		}
	}

	if (i == g_nr_cmds) {
		fprintf(stderr, "Unknown command '%s'\n", argv[0]);
		ret = -1;
	}

	close(g_devfd);
	return ret;
}


static void
print_status_bit(u32_t flags, u32_t bit, char *str, char *true_state,
		char *false_state)
{
	if (flags & bit)
		printf("    %s: %s\n", str, true_state);
	else
		printf("    %s: %s\n", str, false_state);
}

static void
print_status_flags(u32_t flags)
{
	print_status_bit(flags, WAM_STAT_DATA_INVALID,
		       "Data Valid\t\t", "No", "Yes");

	print_status_bit(flags, WAM_STAT_DATA_DIRTY,
		       "Data State\t\t", "Dirty", "Clean");

	print_status_bit(flags, WAM_STAT_WRITE_READY,
		       "Write Ready\t\t", "Yes", "No");

	print_status_bit(flags, WAM_STAT_SSD_CONN,
		       "SSD Connected\t", "Yes", "No");

	print_status_bit(flags, WAM_STAT_SCAP_CONN,
		       "Supercap Connected\t", "Yes", "No");

	print_status_bit(flags, WAM_STAT_SCAP_CHARGING,
		       "Supercap Charging\t", "Yes", "No");

	print_status_bit(flags, WAM_STAT_SCAP_DISCHARGE_ON,
		       "Supercap Discharge\t", "On", "Off");

	print_status_bit(flags, WAM_STAT_DMA_ON,
		       "Host DMA\t\t", "Initialized", "Not Initialized");
}

static void
wc_check_sn(char *sn, u32_t size)
{
	u32_t i;

	for (i = 0; i < size; i++) {
		if (isprint(sn[i]))
			continue;
		sn[i] = 0;
		break;
	}
}

static u32_t
print_event_data(char *str, u32_t len, u32_t event, u32_t data)
{
	int ret = 0;
	u32_t kb, mb;

	switch (event) {
	case WAMEV_SCAP_VOL:
	case WAMEV_SCAP_VOL_LOW:
		ret = snprintf(str, len, "%u mV", data);
		break;

	case WAMEV_FW_BOOTUP:
		if (data)
			ret = snprintf(str, len, "Cold Boot");
		else
			ret = snprintf(str, len, "Warm Boot");
		break;

	case WAMEV_DEV_WR_READY:
		ret = snprintf(str, len, "0x%08X", data);
		break;

	case WAMEV_DATA_BACKUP_START:
		mb = (data >> 17);
		kb = ((data >> 7) & ((1 << 10) - 1));
		if (mb)
			ret = snprintf(str, len, "%u MB ", mb);

		if (!mb && kb)
			ret += snprintf(str + ret, len - ret, "%u KB", kb);
		break;

	case WAMEV_POWER_DOWN:
	case WAMEV_POWER_ON:
	case WAMEV_PCIE_RESET:
		ret = snprintf(str, len, "%u mV, 0x%x", data & 0xFFFF,
				data >> 24);
		break;

	default:
		ret = snprintf(str, len, "0x%X", data);
		break;
	}

	return ret;
}

static inline u32_t
print_time_digits(char *str, u32_t len, u32_t digit,
		char sep, u32_t *leading_zero)
{
	if ((*leading_zero) == 0 && digit == 0) {
		return snprintf(str, len, "__%c", sep);
	}

	if (digit)
		(*leading_zero)++;

	return snprintf(str, len, "%02u%c", digit, sep);
}

static void
uptime_to_str(u32_t uptime, char *str, u32_t len)
{
	u32_t tmp;
	u32_t time_unit;
	u32_t count = 0;
	u32_t leading_zero = 0;

	/* secs per year */
	tmp = 60 * 60 * 24 * 365;
	time_unit = uptime / tmp;

	count += print_time_digits(str + count, len - count,
			time_unit, '-', &leading_zero);
	uptime %= tmp;

	/* secs per month */
	tmp = 60 * 60 * 24 * 30;
	time_unit = uptime / tmp;

	count += print_time_digits(str + count, len - count,
			time_unit, '-', &leading_zero);
	uptime %= tmp;

	/* secs per day */
	tmp = 60 * 60 * 24;
	time_unit = uptime / tmp;

	count += print_time_digits(str + count, len - count,
			time_unit, ' ', &leading_zero);
	uptime %= tmp;


	/* secs per hour */
	tmp = 60 * 60;
	time_unit = uptime / tmp;

	count += print_time_digits(str + count, len - count,
			time_unit, ':', &leading_zero);
	uptime %= tmp;

	/* secs per min */
	tmp = 60;
	time_unit = uptime / tmp;

	count += print_time_digits(str + count, len - count,
			time_unit, ':', &leading_zero);
	uptime %= tmp;

	leading_zero = 1;
	count += print_time_digits(str + count, len - count,
			uptime, ' ', &leading_zero);
}

static int
wc_backup_log(int fd)
{
	int ret;
	backup_stats_t	stats;
	u64_t ddr_offset;
	int i;

	memset(&stats, 0, sizeof(backup_stats_t));

	ret = ioctl(fd, NVRAM_IOC_GET_BACKUP_STATS, &stats);
	if (ret != 0) {
		fprintf(stderr, "Get Backup Stats failed: %d\n", ret);
		perror("ioctl");
		return -1;
	}

	ddr_offset = stats.b_ddr_offset;
	ddr_offset <<= 3;

	printf("\nBackup Log:\n");

	printf("    Status\t\t: %s\n",
			stats.b_completed ? "Completed" : "INCOMPLETE");
	printf("    IO Error count\t: %u\n", stats.b_err_count);
	printf("    DRAM Offset\t\t: %llu MB\n", ddr_offset >> 20);

	for (i = 0; i < 4; i++) {
		printf("    Progress %2u%%\t: %u s\tMin IO Time: %u us\t"
				"Max IO Time: %u us\n", i * 25,
				stats.b_start_time[i],
				stats.b_min_time[i],
				stats.b_max_time[i]);
	}

	printf("    SSD WC Flush\t: %u s\n", stats.b_ssd_flush_time);
	printf("    Backup End\t\t: %u s\n", stats.b_end_time);
	printf("    Supercap Lasted\t: %u s\n", stats.b_alive_time);
	printf("    Supercap Start Vol\t: %u mV\n", stats.b_start_scap_vol);
	printf("    Supercap End Vol\t: %u mV\n", stats.b_end_scap_vol);
	printf("    DRAM SBE Count\t: %u\n", stats.b_sbe_cnt);
	printf("    DRAM DBE Count\t: %u\n", stats.b_dbe_cnt);

	printf("\n");

	return 0;
}

static inline void
print_scap_health(u32_t health)
{
	/* very good - if health above 90% */
	printf("    Health\t\t: ");
	if (health > 90)
		printf("Very Good\n");
	else if (health > 80)
		printf("Good\n");
	else if (health > 75)
		printf("Fair\n");
	else
		printf("Poor\n");
}


static int
wc_info(int argc, char **argv)
{
	int		ret;
	char		str[256];
	u32_t		tmp;
	wam_info_t	info;
	wam_status_t	*dev_stat;
	u64_t		status;
	u32_t		verbose = 0;
	int		b_major = 0;
	int		b_minor = 0;
	wam_drv_version_t	ver;

	if (argc >= 1) {
		if (argv[0][0] == 'v') {
			verbose = 1;
		} else {
			fprintf(stderr, "Unknown option '%s'\n", argv[0]);
			return -1;
		}
	}

	memset(&info, 0, sizeof(wam_info_t));
	memset(&ver, 0, sizeof(wam_drv_version_t));

	ret = ioctl(g_devfd, NVRAM_IOC_DRV_VERSION, &ver);
	if (ret != 0) {
		fprintf(stderr, "DRV_VERSION call failed!\n");
		perror("ioctl");
	}

	wam_get_board_id(g_devfd, &b_major, &b_minor);

	ret = ioctl(g_devfd, NVRAM_IOC_GET_INFO, &info);
	if (ret != 0) {
		fprintf(stderr, "Get dev info call failed!\n");
		perror("ioctl");
		return -1;
	}

	ret = ioctl(g_devfd, NVRAM_IOC_WRITE_READY, &status);
	if (ret != 0) {
		fprintf(stderr, "Get dev status failed!\n");
		perror("ioctl");
		return -1;
	}
	dev_stat = (wam_status_t *)&status;

	wc_check_sn(info.w_sn, WAM_MAX_SN_LEN);
	info.w_sn[WAM_MAX_SN_LEN - 1] = 0;
	printf("Serial Numbers:\n");
	printf("    Device\t\t: %s\n", info.w_sn);

	if (b_major < 3) {
		memcpy(str, info.w_ssd_sn, WAM_SSD_SN_LEN);
		str[WAM_SSD_SN_LEN] = 0;
		printf("    SSD\t\t\t: %s\n", str);
	}

	printf("\n");

	printf("Versions:\n");
	printf("    Driver\t\t: %u.%u.%u\n", ver.w_drv_major, ver.w_drv_minor,
			ver.w_drv_bugfix);
	printf("    Firmware\t\t: %s\n", info.w_fw_ver);

	memcpy(str, info.w_ssd_fw_ver, WAM_SSD_FW_VER_LEN);
	str[WAM_SSD_FW_VER_LEN] = 0;
	printf("    SSD Firmware\t: %s\n", str);

	printf("    EEPROM\t\t: %u\n", info.w_eeprom_ver);
	printf("    DDR Config\t\t: %u\n", info.w_ddr_cfg_ver);
	printf("    Board Rev\t\t: %u.%u\n", b_major, b_minor);

	printf("\n");

	printf("Clock Settings:\n");
	printf("    DDR\t\t\t: %u MHz\n", info.w_dclk);
	printf("    Internal Bus\t: %u MHz\n", info.w_tclk);
	printf("    SoC CPU\t\t: %u MHz\n", info.w_pclk);

	printf("\n");

	uptime_to_str(info.w_uptime, str, 256);
	printf("Firmware Uptime\t\t: %s\n\n", str);

	printf("Status Flags\t\t: %08X\n", info.w_flags);
	print_status_flags(info.w_flags);

	/* show backup error log in case of error */
	if (info.w_flags & WAM_STAT_DATA_INVALID)
		verbose = 1;

	printf("\n");
	printf("Supercap Info:\n");
	printf("    Status\t\t: %s\n", scap_status[dev_stat->w_scap_stat]);
	printf("    Current charge\t: %u mV (%u %%)\n",
			info.w_sc_vol, dev_stat->w_scap_charge);
	printf("    Expected charge\t: %u mV\n", info.w_sc_max_vol);

	tmp = info.w_sc_cap * 10000;
	tmp /= info.w_sc_design_cap;
	if (info.w_sc_cap) {
		printf("    Capacitance\t\t: %u F (%u.%u %%)\n",
				info.w_sc_cap, tmp / 100, tmp % 100);
		print_scap_health(tmp / 100);
	} else {
		printf("    Capacitance\t\t: Health Check Not Done\n");
	}
	printf("    Design Cap\t\t: %u F\n", info.w_sc_design_cap);

	if (info.w_flags & WAM_STAT_SCAP_CALIB_ON) {
		printf("    Health Check\t: In Progress (%u %% complete)\n",
				info.w_sc_calib_progress);
	}

	printf("\n");
	printf("DRAM Info\n");
	printf("    Size\t\t: %u KB\n", info.w_dram_size);
	printf("    Usable Size\t\t: %u KB\n",
			info.w_dram_size - info.w_dram_rsvd_size);

	printf("    Memory Test\t\t: %s\n",
			memtest_status[dev_stat->w_post_res]);
	printf("    Single Bit Errors\t: %u\n", info.w_dram_sbe);
	printf("    Multi Bit Errors\t: %u\n", info.w_dram_dbe);

	printf("\n");
	printf("SSD Info:\n");
	printf("    Status\t\t: %s\n", ssd_status[dev_stat->w_ssd_stat]);
	printf("    Average Erase Count\t: %u\n", info.w_ssd_avg_erase_cnt);
	printf("    Program Error Count\t: %u\n", info.w_ssd_prg_err_cnt);
	printf("    Erase Error Count\t: %u\n", info.w_ssd_erase_err_cnt);
	printf("    Defective Blocks\t: %u (8MB each)\n",
			info.w_ssd_defect_blks);

	printf("\n");
	printf("Unread Events\t\t: %u\n", info.w_events_unread);
	printf("Board Temperature\t: %u C\n", info.w_board_temp);

	if (verbose)
		wc_backup_log(g_devfd);

	return 0;
}

static int
wc_clear_events(u32_t rd_ptr)
{
	/* Set new read ptr */
	if (ioctl(g_devfd, NVRAM_IOC_SET_EVNTLOG_RDPTR, rd_ptr) < 0) {
		fprintf(stderr, "set event log read ptr failed\n");
		return -1;
	}
	return 0;
}

static int
wc_events(int argc, char **argv)
{
	char *addr;
	char str[256];
	wam_event_t *events;
	event_log_t e_log;
	u32_t rd_ptr, wr_ptr, max_e;
	int clear = 0;

	if (argc >= 1) {
		if (strcasecmp(argv[0], "clear") == 0) {
			clear = 1;
		} else {
			fprintf(stderr, "Unknown option '%s'\n", argv[0]);
			return -1;
		}
	}

	/* Get Event Log Parameters */
	if (ioctl(g_devfd, NVRAM_IOC_GET_EVNTLOG_PTRS, &e_log) < 0) {
		fprintf(stderr, "get event log parameters failed\n");
		perror("ioctl");
		return -1;
	}

	rd_ptr = e_log.e_rd_ptr;
	wr_ptr = e_log.e_wr_ptr;
	max_e  = e_log.e_max_events;

	if ((rd_ptr >= max_e) || (wr_ptr >= max_e)) {
		fprintf(stderr, "Device event log seems to be corrupt! "
			       "[%08x|%08x|%08x]\n",
				rd_ptr, wr_ptr, max_e);
		return -1;
	}

	if(rd_ptr == wr_ptr) {
		printf("No new event\n");
		return 0;
	}

	addr = mmap(0, NV_NVSRAM_ACCESS_SIZE, PROT_READ | PROT_WRITE,
			MAP_SHARED, g_devfd, NV_MMAP_NVSRAM_OFFSET);
	if (addr == MAP_FAILED) {
		fprintf(stderr, "mmap error\n");
		return -1;
	}

	events = (wam_event_t *)(addr + e_log.e_nvsram_offset);

	do {
		rd_ptr++;
		if(rd_ptr >= max_e)
			rd_ptr = 0;

		uptime_to_str(events[rd_ptr].e_time, str, 256);
		printf("[%s]: ", str);

		if (events[rd_ptr].e_num >= WAMEV_LAST_ENTRY) {
			snprintf(str, 64, "Unknown Event [%u]",
					events[rd_ptr].e_num);
			printf("%-40s - ", str);
		} else {
			printf("%-40s - ", event_str[events[rd_ptr].e_num]);
		}

		print_event_data(str, 256, events[rd_ptr].e_num,
				events[rd_ptr].e_data);
		printf("%s\n", str);

	} while (rd_ptr != wr_ptr);

	if (clear)
		wc_clear_events(rd_ptr);

	munmap(addr, NV_NVSRAM_ACCESS_SIZE);
	return 0;
}

static int
wc_ping(int argc, char **argv)
{
	int ret;

	ret = ioctl(g_devfd, NVRAM_IOC_PING_FW, 0);
	if (ret != 0) {
		fprintf(stderr, "Firmware ping failed\n");
		perror("ioctl");
		return -1;
	}

	printf("Firmware responded with Hello!\n");
	return 0;
}

/*
 * Firmware reset
 */
static int
wc_reset(int argc, char **argv)
{
	int ret;

	ret = ioctl(g_devfd, NVRAM_IOC_RESET_FW, 0);
	if (ret != 0) {
		fprintf(stderr, "Resetting device failed: %d\n", ret);
		perror("ioctl");
		return -1;
	}

	printf("Firmware Reset successful\n");
	return 0;
}

/*
 * Hard reset
 */
static int
wc_hardreset(int argc, char **argv)
{
	int ret;

	ret = ioctl(g_devfd, NVRAM_IOC_HARDRESET, 0);
	if (ret != 0) {
		fprintf(stderr, "Hard resetting failed: %d\n", ret);
		perror("ioctl");
		return -1;
	}

	printf("Hard Reset successful\n");
	return 0;
}

/*
 * Format the SSD
 */
static int
wc_ssd_format(int argc, char **argv)
{
	int ret;

	ret = ioctl(g_devfd, NVRAM_IOC_FORMAT_SSD, 0);
	if (ret != 0) {
		fprintf(stderr, "SSD format failed: %d\n", ret);
		perror("ioctl");
		return -1;
	}

	printf("SSD DOM formatted\n");
	return 0;
}

static int
wc_list_errs(int argc, char **argv)
{
	int ret;
	err_block_t *eblks;
	u64_t ddr_offset;
	int i;

	eblks = (err_block_t *) malloc(4096);
	memset(eblks, 0, 4096);

	ret = ioctl(g_devfd, NVRAM_IOC_GET_ERR_BLOCKS, eblks);
	if (ret != 0) {
		fprintf(stderr, "Get Err Blocks ioctl failed: %d\n", ret);
		perror("ioctl");
		return -1;
	}

	i = 0;
	while (eblks[i].e_blk_size != 0 || eblks[i].e_qword_off != 0) {
		ddr_offset = eblks[i].e_qword_off;
		ddr_offset <<= 3;
		printf("[%u] offset: 0x%010llx size: 0x%08x, op: %s\n", i,
				ddr_offset, eblks[i].e_blk_size,
				error_type[eblks[i].e_err_op]);
		i++;
	}

	printf("Device recorded %u errors\n", i);
	return 0;
}


static int
wc_acquit(int argc, char **argv)
{
	int ret;

	ret = ioctl(g_devfd, NVRAM_IOC_CLEAR_BACKUP_STATS, 0);
	if (ret != 0) {
		fprintf(stderr, "Clear Device Error ioctl failed: %d\n", ret);
		perror("ioctl");
		return -1;
	}

	printf("Backup/Restore error cleared\n");

	return 0;
}

static int
wc_ssd_update_raw(int argc, char **argv)
{
	int		ret;
	int		fw_fd;
	struct stat	filestat;
	unsigned long	fsize;
	char		*srcbuf;
	file_info_t	file_info;

	memset(&file_info, 0, sizeof(file_info_t));

	fw_fd = open(argv[0], O_RDONLY);
	if (fw_fd < 0) {
		fprintf(stderr, "Failed to open file %s\n", argv[0]);
		perror("open");
		return -1;
	}

	if (fstat(fw_fd, &filestat) != 0) {
		fprintf(stderr, "Failed to fstat file %s\n", argv[0]);
		perror("fstat");
		return -1;
	}

	fsize = filestat.st_size;
	srcbuf = (char *)mmap(NULL, fsize, PROT_READ, MAP_PRIVATE, fw_fd, 0);
	if (srcbuf == MAP_FAILED) {
		fprintf(stderr, "Failed to mmap ssd firmware file\n");
		perror("mmap");
		return -1;
	}

	file_info.file_buf = srcbuf;
	file_info.file_size = fsize;

	ret = ioctl(g_devfd, NVRAM_IOC_SSD_FW_UPDATE, &file_info);
	if (ret != 0) {
		fprintf(stderr, "Update SSD FW ioctl failed: %d\n", ret);
		perror("ioctl");
		return -1;
	}

	munmap(srcbuf, fsize);
	close(fw_fd);
	return 0;
}

static int
wc_update_eeprom_raw(int argc, char **argv)
{
	int		ret;
	int		eeprom_fd;
	struct stat	filestat;
	unsigned long	fsize;
	char		*srcbuf;
	file_info_t	file_info;

	memset(&file_info, 0, sizeof(file_info_t));

	eeprom_fd = open(argv[0], O_RDONLY);
	if (eeprom_fd < 0) {
		fprintf(stderr, "Failed to open file %s\n", argv[0]);
		perror("open");
		return -1;
	}

	if (fstat(eeprom_fd, &filestat) != 0) {
		fprintf(stderr, "Failed to fstat file %s\n", argv[0]);
		perror("fstat");
		close(eeprom_fd);
		return -1;
	}

	fsize = filestat.st_size;
	if (fsize > MV_EEPROM_SIZE) {
		printf("File size (%lu) too large for eeprom. Max size: %u\n",
				fsize, MV_EEPROM_SIZE);
		close(eeprom_fd);
		return -1;
	}

	srcbuf = mmap(NULL, fsize, PROT_READ, MAP_PRIVATE, eeprom_fd, 0);
	if (srcbuf == MAP_FAILED) {
		fprintf(stderr, "Failed to mmap eeprom file\n");
		perror("mmap");
		close(eeprom_fd);
		return -1;
	}

	file_info.file_buf = srcbuf;
	file_info.file_size = fsize;

	ret = ioctl(g_devfd, NVRAM_IOC_UPDATE_EEPROM, &file_info);
	if (ret != 0)
		printf("EEPROM Updated failed! err: %d\n", ret);

	munmap(srcbuf, fsize);
	close(eeprom_fd);

	return ret;
}

static int
wc_read_eeprom(int argc, char **argv)
{
	int		ret;
	u8_t		*buf;
	file_info_t	file_info;

	memset(&file_info, 0, sizeof(file_info_t));

	buf = (u8_t *) malloc(MV_EEPROM_SIZE);
	if (buf == NULL) {
		fprintf(stderr, "Failed to allocate memory\n");
		return -1;
	}

	file_info.file_buf = buf;
	file_info.file_size = MV_EEPROM_SIZE;

	ret = ioctl(g_devfd, NVRAM_IOC_READ_EEPROM, &file_info);
	if (ret != 0) {
		fprintf(stderr, "EEPROM read failed! err: %d\n", ret);
		free(buf);
		return -1;
	}

	ret = write(1, buf, MV_EEPROM_SIZE);

	if (ret != MV_EEPROM_SIZE) {
		fprintf(stderr, "write to stdout failed!\n");
		perror("write");
	}

	return ret;
}

static void
dump_reg_range(void *addr, u32_t start, u32_t end)
{
	u32_t i;

	for (i = start; i <= end; i += 4)
		printf("%05X: %08X\n", i, *(u32_t *)(addr + i));

	printf("\n");
}


static int
wc_regdump(int argc, char **argv)
{
	char *addr;

	/* memory map flash area */
	addr = mmap (0, NV_REG_SPACE_SIZE, PROT_READ | PROT_WRITE,
			MAP_SHARED, g_devfd, NV_MMAP_REG_OFFSET);
	if (addr == MAP_FAILED) {
		fprintf(stderr, "mmap of device memory failed\n");
		perror("mmap");
		return -1;
	}

	dump_reg_range(addr, 0x10010, 0x10010);
	dump_reg_range(addr, 0x10100, 0x1011c);
	dump_reg_range(addr, 0x1045c, 0x104e8);
	dump_reg_range(addr, 0x11010, 0x1105c);
	dump_reg_range(addr, 0x12000, 0x12100);
	dump_reg_range(addr, 0x20000, 0x20080);
	dump_reg_range(addr, 0x20100, 0x20114);
	dump_reg_range(addr, 0x20200, 0x20210);
	dump_reg_range(addr, 0x20300, 0x20324);
	dump_reg_range(addr, 0x20400, 0x2044c);
	dump_reg_range(addr, 0x20500, 0x2050c);
	dump_reg_range(addr, 0x20544, 0x2054c);
	dump_reg_range(addr, 0x31804, 0x31808);
	dump_reg_range(addr, 0x31820, 0x318c4);
	dump_reg_range(addr, 0x31820, 0x318c4);
	dump_reg_range(addr, 0x31900, 0x31910);
	dump_reg_range(addr, 0x31A00, 0x31abc);
	dump_reg_range(addr, 0x31b00, 0x31b24);
	dump_reg_range(addr, 0x80000, 0x801fc);
	dump_reg_range(addr, 0xf1400, 0xf1628);
	dump_reg_range(addr, 0x900, 0x998);
	dump_reg_range(addr, 0x70900, 0x70998);
	dump_reg_range(addr, 0x50000, 0x50004);
	dump_reg_range(addr, 0x60900, 0x60998);
	dump_reg_range(addr, 0x61900, 0x61998);

	munmap(addr, NV_REG_SPACE_SIZE);

	return 0;
}

static int
wc_scap_calib(int argc, char **argv)
{
	int ret;
	int major, minor;
	u32_t days = 0;

	if (argc == 1) {
		days = strtoull(argv[0], NULL, 10);
		if (days < 1 || days > 30) {
			printf("Invalid argument: %s. Valid schedule in days: "
					"1 to 30\n", argv[0]);
			return (-1);
		}
	}

	ret = wam_get_board_id(g_devfd, &major, &minor);
	if (ret != 0) {
		printf("Unable to determine board revision. "
				"Cannot start health measurement\n");
		return ret;
	}

	if (major < 3 || minor < 3) {
		printf("Supercap health measurement not supported for "
				"board revision %d.%d\n", major, minor);
		return (0);
	}

	ret = ioctl(g_devfd, NVRAM_IOC_SCAP_CALIB_START, days);
	if (ret != 0) {
		if (errno == EBUSY) {
			fprintf(stderr, "Supercap health check already "
					"in progress.\n");
			return 0;
		} else {
			fprintf(stderr, "Start supercap health check "
					"failed: %d\n", ret);
			perror("ioctl");
			return ret;
		}
	}

	printf("Supercap health check started. Estimated time: 10 to "
		       "30 mins\n");
	printf("Run 'wamctl /dev/mvwam<idx> info' to see the progress\n");
	return 0;
}

static int
wc_reg_rw(int argc, char **argv)
{
	int		ret;
	reg_io_t	reg_cfg;
	char		*next_char;

	memset(&reg_cfg, 0, sizeof (reg_io_t));

	reg_cfg.reg_addr = strtoull(argv[0], &next_char, 16);

	if (*next_char == '=' && *(next_char + 1) != 0) {
		reg_cfg.rw = 1;
		reg_cfg.reg_val = strtoull(next_char + 1, NULL, 16);
	}

	if (argc > 1) {
		if (reg_cfg.rw) {
			fprintf(stderr, "Too many arguments!\n");
			return -1;
		}
		reg_cfg.rw = 1;
		reg_cfg.reg_val = strtoull(argv[1], NULL, 16);
	}

	if (reg_cfg.reg_addr > (1 << 20)) {
		fprintf(stderr, "Invalid address: %X. Addressable range: "
				"0 to %X\n", reg_cfg.reg_addr, (1 << 20) - 1);
		return -1;
	}

	ret = ioctl(g_devfd, NVRAM_IOC_REG_IO, &reg_cfg);

	if (ret == 0)
		printf("%08X: %08X\n", reg_cfg.reg_addr, reg_cfg.reg_val);
	else
		perror("ioctl");

	return ret;
}

static wc_update_cmd_t wc_update_cmd_tbl[] = {
	{
		.u_long_cmd	= "firmware",
		.u_short_cmd	= "fw",
		.u_update_fn	= wam_update_firmware
	},
	{
		.u_long_cmd	= "ssd_firmware",
		.u_short_cmd	= "ssdfw",
		.u_update_fn	= wam_update_ssdfw
	},
	{
		.u_long_cmd	= "eeprom",
		.u_short_cmd	= "ep",
		.u_update_fn	= wam_update_eeprom
	},
	{
		.u_long_cmd	= "bootloader",
		.u_short_cmd	= "bl",
		.u_update_fn	= wam_update_uboot
	},
	{
		.u_long_cmd	= "device_config",
		.u_short_cmd	= "cfg",
		.u_update_fn	= wam_update_ddr_cfg
	},
	{
		.u_long_cmd	= "package",
		.u_short_cmd	= "pkg",
		.u_update_fn	= wam_update_package
	},
};


static wc_update_cmd_t *
wc_update_parse_cmd(char *cmd_str, char **parse_filename)
{
	int			i;
	char			*arg, *cmd;

	assert(cmd_str && parse_filename);

	cmd = strtok(cmd_str, "=");
	arg = strtok(NULL, "=");

	for (i = 0; i < NUM_OF_ELM(wc_update_cmd_tbl); i++) {
		if ((strcmp(cmd, wc_update_cmd_tbl[i].u_long_cmd) == 0) ||
		(strcmp(cmd, wc_update_cmd_tbl[i].u_short_cmd) == 0)) {
			*parse_filename = arg;
			return (&wc_update_cmd_tbl[i]);
		}
	}

	*parse_filename = arg;
	return (NULL);
}

static void
print_hidden_update_usage(void)
{
	printf("Usage:\tpackage|pkg=<package file>\t\tOR\n"
			"\tfirmware|fw=<firmware file>\t\tOR\n"
			"\tssd_firmware|ssdfw=<ssd firmware file>\tOR\n"
			"\teeprom|ep=<eeprom image file>\t\tOR\n"
			"\tbootloader|bl=<boot loader file>\tOR\n"
			"\tdevice_config|cfg=<device config file>\n");
}

/**
 * wc_update:
 * -----------
 *
 * Update firmware, ssd_firmware, eeprom, bootloader, and device_config
 *
 *
 * .usage		= "firmware|fw=<firmware file> "
 *				"ssd_firmware|ssdfw=<ssd firmware file>"
 *				" eeprom|ep=<eeprom image file> "
 *				"bootloader|bl=<boot loader file> "
 *				"device_config|cfg=<device config file>",
 */
static int
wc_update(int argc, char **argv)
{
	int			ret;
	int			force = 0;
	char			*update_file = NULL;
	wc_update_cmd_t	*update_cmd = NULL;

	if (strcasecmp(argv[0], WAM_HIDDEN_HELP_STR) == 0) {
		print_hidden_update_usage();
		return (0);
	}

	update_cmd = wc_update_parse_cmd(argv[0], &update_file);
	if (update_cmd == NULL || update_file == NULL) {
		fprintf(stderr, "Invalid command option: %s\n", argv[0]);
		return (-1);
	}

	if (argc > 1) {
		if ((strcasecmp(argv[1], "-f") == 0) ||
				(strcasecmp(argv[1], "-force") == 0)) {
			force = 1;
		} else {
			fprintf(stderr, "Unknown option %s\n", argv[1]);
			return (-1);
		}
	}

	ret = update_cmd->u_update_fn(g_devfd, update_file, force);

	return (ret);
}

static int
wc_board_id(int argc, char **argv)
{
	int		major, minor;

	if (wam_get_board_id(g_devfd, &major, &minor) == 0) {
		printf("%d.%d\n", major, minor);
		return (0);
	} else {
		printf("Unable to determine board ID\n");
		return (-1);
	}

}

static int
wc_print_vpd(int argc, char **argv)
{
	if ((argc == 1) && (strcmp(argv[0], "old") == 0)) {
		wam_vpd_print(g_devfd);
		printf("---------------------------------------------------\n");
	}
	wam_vpd_v2_print(g_devfd);
	return (0);
}

static int
wc_convert_vpd(int argc, char **argv)
{
	return (wam_convert_vpd(g_devfd));
}

static int
wc_prog_vpd(int argc, char **argv)
{
	assert(argc == 5);
	return (wam_prog_vpd(g_devfd, argc, argv));
}


static int
wc_nvsram_dump(int argc, char **argv)
{
	int		ret;
	u8_t		*buf;

	buf = (u8_t *) malloc(MV_NVSRAM_SIZE);
	if (buf == NULL) {
		fprintf(stderr, "Failed to allocate memory\n");
		return -1;
	}

	ret = wam_nvsram_dump(g_devfd, buf, MV_NVSRAM_SIZE);
	if (ret != 0) {
		fprintf(stderr, "nvsram dump api failed!\n");
		free(buf);
		return ret;
	}

	ret = write(1, buf, MV_NVSRAM_SIZE);
	if (ret != MV_NVSRAM_SIZE) {
		perror("write");
	}

	return !ret;
}

